# -*- coding: utf-8 -*-
# Copyright 2007-2026 The HyperSpy developers
#
# This file is part of RosettaSciIO.
#
# RosettaSciIO is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# RosettaSciIO is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with RosettaSciIO. If not, see <https://www.gnu.org/licenses/#GPL>.
import importlib.util
from pathlib import Path

import numpy as np
import pytest

from rsciio.exceptions import VisibleDeprecationWarning
from rsciio.mrc import file_reader
from rsciio.mrc._api import (
    MOVIE_RE,
    VIRTUAL_RE,
    find_related_de_files,
    get_data_type,
)

hs = pytest.importorskip("hyperspy.api", reason="hyperspy not installed")
zarr_missing = importlib.util.find_spec("zarr") is None


TEST_DATA_DIR = Path(__file__).parent / "data" / "mrc"


def test_single_image():
    # Acquired from Velox
    s = hs.load(TEST_DATA_DIR / "HAADFscan.mrc")
    assert s.data.shape == (16, 16)
    assert s.axes_manager.signal_shape == (16, 16)
    assert s.axes_manager.navigation_shape == ()

    for axis in s.axes_manager.signal_axes:
        assert axis.scale == 5.679131317138672
        assert axis.offset == 0
        assert axis.units == "nm"


def test_4DSTEM_image():
    # Acquired from Velox
    s = hs.load(TEST_DATA_DIR / "4DSTEMscan.mrc")
    assert s.data.shape == (256, 256, 256)
    assert s.axes_manager.signal_shape == (256, 256)
    assert s.axes_manager.navigation_shape == (256,)


def test_4DSTEM_image_navigation_shape_16_16():
    # Acquired from Velox
    s = hs.load(
        TEST_DATA_DIR / "4DSTEMscan.mrc",
        navigation_shape=(16, 16),
    )
    assert s.data.shape == (16, 16, 256, 256)
    assert s.axes_manager.signal_shape == (256, 256)
    assert s.axes_manager.navigation_shape == (16, 16)


def test_4DSTEM_image_navigation_shape_8_32():
    s = hs.load(
        TEST_DATA_DIR / "4DSTEMscan.mrc",
        navigation_shape=(8, 32),
    )
    assert s.data.shape == (32, 8, 256, 256)
    assert s.axes_manager.signal_shape == (256, 256)
    assert s.axes_manager.navigation_shape == (8, 32)


@pytest.mark.parametrize("distributed", [True, False])
def test_distributed_deprecation_warning(distributed):
    with pytest.warns(VisibleDeprecationWarning):
        file_reader(
            str(TEST_DATA_DIR / "4DSTEMscan.mrc"),
            navigation_shape=(8, 32),
            distributed=distributed,
        )


def test_deprecated_mmap_mode():
    with pytest.warns(VisibleDeprecationWarning):
        file_reader(
            str(TEST_DATA_DIR / "4DSTEMscan.mrc"),
            navigation_shape=(8, 32),
            mmap_mode="r",
        )


def test_mrc_chunks_equal():
    s = hs.load(
        TEST_DATA_DIR / "4DSTEMscan.mrc",
        navigation_shape=(8, 32),
        chunks=(16, 4, 256, 256),
        lazy=True,
    )
    assert s.data.chunks == (
        (
            16,
            16,
        ),
        (
            4,
            4,
        ),
        (256,),
        (256,),
    )


@pytest.mark.parametrize("navigation_shape", [None, (256,), (8, 32), (8, 16, 2)])
def test_mrc_metadata(navigation_shape):
    s = hs.load(
        TEST_DATA_DIR / "4DSTEMscan.mrc",
        metadata_file=TEST_DATA_DIR / "info.txt",
        navigation_shape=navigation_shape,
    )
    if navigation_shape is None:
        navigation_shape = (8, 32)
    shape = navigation_shape[::-1] + (256, 256)
    assert s.data.shape == shape
    assert s.axes_manager.signal_shape == (256, 256)
    assert s.axes_manager.navigation_shape == navigation_shape
    assert s.metadata.Acquisition_instrument.TEM.detector == "CeleritasXS"
    assert s.metadata.Acquisition_instrument.TEM.magnification == "1000"
    assert s.metadata.Acquisition_instrument.TEM.frames_per_second == "40000"


def test_mrc_metadata_auto():
    s = hs.load(TEST_DATA_DIR / "20241021_00405_movie.mrc", lazy=True)
    navigation_shape = (8, 4)
    shape = navigation_shape[::-1] + (4, 8)
    assert s.data.shape == shape
    assert s.axes_manager.signal_shape == (8, 4)
    assert s.axes_manager.navigation_shape == navigation_shape
    assert s.metadata.Acquisition_instrument.TEM.detector == "DESim"
    assert s.metadata.Acquisition_instrument.TEM.magnification == "1000"
    assert s.metadata.Acquisition_instrument.TEM.frames_per_second == "700"
    assert len(s.metadata._HyperSpy.navigators.keys()) == 3
    assert isinstance(s.metadata._HyperSpy.navigators["Virt 0"], hs.signals.Signal2D)

    assert s.metadata._HyperSpy.navigator is not None

    shape = (
        s.axes_manager._navigation_shape_in_array
        + s.axes_manager._signal_shape_in_array
    )
    assert s.data.shape == shape


def test_mrc_metadata_auto_custom_shape():
    s = hs.load(
        TEST_DATA_DIR / "20241021_00405_movie.mrc", lazy=True, navigation_shape=(16, 2)
    )
    navigation_shape = (16, 2)
    shape = navigation_shape[::-1] + (4, 8)
    assert s.data.shape == shape
    assert s.axes_manager.signal_shape == (8, 4)
    assert s.axes_manager.navigation_shape == navigation_shape
    assert s.metadata.Acquisition_instrument.TEM.detector == "DESim"
    assert s.metadata.Acquisition_instrument.TEM.magnification == "1000"
    assert s.metadata.Acquisition_instrument.TEM.frames_per_second == "700"
    assert len(s.metadata._HyperSpy.navigators) == 3

    assert s.metadata._HyperSpy.navigator is not None
    assert s.metadata._HyperSpy.navigator.data.shape == navigation_shape[::-1]

    assert isinstance(s.metadata._HyperSpy.navigator, hs.signals.BaseSignal)

    shape = (
        s.axes_manager._navigation_shape_in_array
        + s.axes_manager._signal_shape_in_array
    )
    assert s.data.shape == shape


@pytest.mark.parametrize(
    "metadata_file",
    [
        TEST_DATA_DIR / "3DSTEM_scan_info.txt",
        TEST_DATA_DIR / "3DTEM_scan_info.txt",
        TEST_DATA_DIR / "3DTEMDiffracting_scan_info.txt",
    ],
)
def test_mrc_metadata_modes(metadata_file):
    s = hs.load(TEST_DATA_DIR / "20241021_00405_movie.mrc", metadata_file=metadata_file)
    diffracting = "STEM" in metadata_file.name or "Diffracting" in metadata_file.name
    s.axes_manager.navigation_axes[0].units = "sec"
    if diffracting:
        s.axes_manager.signal_axes[0].units = "nm^-1"
        s.axes_manager.signal_axes[1].units = "nm^-1"
    else:
        s.axes_manager.signal_axes[0].units = "nm"
        s.axes_manager.signal_axes[1].units = "nm"


def test_mrc_random_scan_pattern():
    s = hs.load(
        TEST_DATA_DIR / "ROI_Random_Scan_movie.mrc",
        metadata_file=TEST_DATA_DIR / "ROI_Random_Scan_info.txt",
        scan_file=TEST_DATA_DIR / "ROI_Random_Scan_scan_coordinates.csv",
    )
    assert s.data.shape == (29, 12, 16, 16)
    # check to make sure that the Sum image from DE Server matches the sum.
    sum_nav = hs.load(TEST_DATA_DIR / "ROI_Random_Scan_Sum.mrc")
    np.testing.assert_array_almost_equal(s.sum(axis=(2, 3)).data, sum_nav, decimal=-1)


def test_repeated_mrc_custom():
    s = hs.load(
        TEST_DATA_DIR / "Custom_movie.mrc",
        metadata_file=TEST_DATA_DIR / "Custom_info.txt",
        scan_file=TEST_DATA_DIR / "Custom_scan_coordinates.csv",
    )
    assert s.data.shape == (5, 5, 2, 16, 16)
    # make sure that the first and second dataset aren't equal
    assert not np.array_equal(s.data[:, :, 0], s.data[:, :, 1])
    np.testing.assert_array_equal(s.data[:, 1], 0)  # Skipped rows
    np.testing.assert_array_equal(s.data[:, 3], 0)  # Skipped rows


def test_repeated_mrc_custom_error():
    with pytest.raises(ValueError):
        hs.load(
            TEST_DATA_DIR / "Custom_movie.mrc",
            metadata_file=TEST_DATA_DIR / "Custom_info.txt",
            scan_file=TEST_DATA_DIR / "Custom_scan_coordinates.csv",
            chunks=(5, 5, 2, 2, 2),
        )


def test_repeated_mrc_custom_no_scan_file():
    with pytest.raises(ValueError):
        hs.load(
            TEST_DATA_DIR / "Custom_movie.mrc",
            metadata_file=TEST_DATA_DIR / "Custom_info.txt",
        )


# ---------------------------------------------------------------------------
# Tests for new PR code: regex patterns, find_related_de_files, uint8 mode 0
# ---------------------------------------------------------------------------


class TestMovieRegex:
    """Tests for the MOVIE_RE regex pattern."""

    @pytest.mark.parametrize(
        "filename, expected",
        [
            (
                "20241021_00405_movie.mrc",
                {
                    "timestamp": "20241021",
                    "acquisitionNumber": "00405",
                    "optionalSuffix": None,
                    "movieNum": None,
                },
            ),
            (
                "20241021_00405_suffix_movie.mrc",
                {
                    "timestamp": "20241021",
                    "acquisitionNumber": "00405",
                    "optionalSuffix": "suffix",
                    "movieNum": None,
                },
            ),
            (
                "20241021_00405_suffix_extra_movie.mrc",
                {
                    "timestamp": "20241021",
                    "acquisitionNumber": "00405",
                    "optionalSuffix": "suffix_extra",
                    "movieNum": None,
                },
            ),
        ],
    )
    def test_movie_re_matches(self, filename, expected):
        m = MOVIE_RE.fullmatch(filename)
        assert m is not None
        for key, val in expected.items():
            assert m.group(key) == val

    def test_movie_re_no_match(self):
        assert MOVIE_RE.fullmatch("20241021_00405_info.txt") is None
        assert MOVIE_RE.fullmatch("not_a_movie.mrc") is None
        assert MOVIE_RE.fullmatch("20241021_00405_0_Virt 0_sum.mrc") is None


class TestVirtualRegex:
    """Tests for the VIRTUAL_RE regex pattern."""

    @pytest.mark.parametrize(
        "filename, expected_name, expected_calc",
        [
            ("20241021_00405_0_Virt 0_sum.mrc", "Virt 0", "sum"),
            ("20241021_00405_1_Virt 1_sum.mrc", "Virt 1", "sum"),
            (
                "20251103_23389_3_CentroidAmp_centroid_amplitude.mrc",
                "CentroidAmp",
                "centroid_amplitude",
            ),
        ],
    )
    def test_virtual_re_matches(self, filename, expected_name, expected_calc):
        m = VIRTUAL_RE.fullmatch(filename)
        assert m is not None, f"VIRTUAL_RE should match {filename!r}"
        assert m.group("name") == expected_name
        assert m.group("calculationType") == expected_calc

    def test_virtual_re_no_match_movie(self):
        assert VIRTUAL_RE.fullmatch("20241021_00405_movie.mrc") is None

    def test_virtual_re_virt_prefix(self):
        # virtualImageNum with 'virt<digits>' prefix should also match
        m = VIRTUAL_RE.fullmatch("20241021_00405_virt0_BF_sum.mrc")
        assert m is not None
        assert m.group("virtualImageNum") == "virt0"
        assert m.group("name") == "BF"


class TestGetDataType:
    """Tests for the get_data_type function."""

    def test_mode_0_is_uint8(self):
        dtype = get_data_type(np.array([0]))
        assert dtype == np.dtype(np.uint8)

    def test_mode_1_is_int16(self):
        assert get_data_type(np.array([1])) == np.dtype(np.int16)

    def test_mode_2_is_float32(self):
        assert get_data_type(np.array([2])) == np.dtype(np.float32)

    def test_mode_6_is_uint16(self):
        assert get_data_type(np.array([6])) == np.dtype(np.uint16)

    def test_mode_12_is_float16(self):
        assert get_data_type(np.array([12])) == np.dtype(np.float16)

    def test_invalid_mode_raises(self):
        with pytest.raises(ValueError, match="Unrecognised mode"):
            get_data_type(np.array([99]))


class TestFindRelatedDeFiles:
    """Tests for find_related_de_files."""

    def test_find_related_de_files_basic(self):
        movie = str(TEST_DATA_DIR / "20241021_00405_movie.mrc")
        result = find_related_de_files(movie)

        assert result["info_file"] is not None
        assert "20241021_00405_info.txt" in result["info_file"]

        assert len(result["virtual_images"]) == 2
        assert len(result["virtual_image_names"]) == 2
        assert "Virt 0" in result["virtual_image_names"]
        assert "Virt 1" in result["virtual_image_names"]

        assert len(result["external_images"]) == 1
        assert len(result["external_image_names"]) == 1
        assert "Ext 1" in result["external_image_names"]

        # scan CSV is not present for this dataset
        assert result["scan_csv"] is None

    def test_find_related_de_files_non_movie_name(self):
        """Non-matching filenames should return empty result."""
        result = find_related_de_files("not_a_real_movie_file.mrc")
        assert result["info_file"] is None
        assert result["virtual_images"] == []
        assert result["virtual_image_names"] == []
        assert result["external_images"] == []
        assert result["scan_csv"] is None

    def test_find_related_de_files_virtual_images_sorted(self):
        """Virtual image list should be sorted."""
        movie = str(TEST_DATA_DIR / "20241021_00405_movie.mrc")
        result = find_related_de_files(movie)
        assert result["virtual_images"] == sorted(result["virtual_images"])

    def test_find_related_de_files_dedupe(self, tmp_path):
        """_dedupe should rename duplicate virtual image names."""
        import shutil

        # Create a minimal movie file and two virtual images with the same name
        src_movie = TEST_DATA_DIR / "20241021_00405_movie.mrc"
        ts = "20260101"
        acq = "99999"
        movie = tmp_path / f"{ts}_{acq}_movie.mrc"
        shutil.copy(src_movie, movie)
        v0 = tmp_path / f"{ts}_{acq}_0_BF_sum.mrc"
        v1 = tmp_path / f"{ts}_{acq}_1_BF_sum.mrc"
        shutil.copy(src_movie, v0)
        shutil.copy(src_movie, v1)
        # Create a dummy info file
        (tmp_path / f"{ts}_{acq}_info.txt").write_text("Camera Model = TestCamera\n")

        result = find_related_de_files(str(movie))
        names = result["virtual_image_names"]
        assert len(names) == 2
        # First occurrence keeps the name, second gets a suffix
        assert names[0] == "BF"
        assert names[1] != "BF"
        assert names[1].startswith("BF")


class TestMetadataVirtualImageNames:
    """Test that virtual and external image names appear correctly in loaded signal metadata."""

    def test_virtual_image_names_in_metadata(self):
        s = hs.load(TEST_DATA_DIR / "20241021_00405_movie.mrc", lazy=True)
        # HyperSpy stores _sig_* keys and exposes them without the prefix as signals
        keys = list(s.metadata._HyperSpy.navigators.as_dictionary().keys())
        assert "_sig_Virt 0" in keys
        assert "_sig_Virt 1" in keys
        # Accessible via HyperSpy attribute (prefix stripped)
        assert "Virt 0" in s.metadata._HyperSpy.navigators

    def test_external_image_names_in_metadata(self):
        s = hs.load(TEST_DATA_DIR / "20241021_00405_movie.mrc", lazy=True)
        nav_keys = list(s.metadata._HyperSpy.navigators.as_dictionary().keys())
        # 2 virtual + 1 external = 3 navigators total
        assert len(nav_keys) == 3
        # External key stored with _sig_ prefix
        assert "_sig_Ext 1" in nav_keys
        # Accessible via HyperSpy attribute (prefix stripped)
        assert "Ext 1" in s.metadata._HyperSpy.navigators

    def test_navigator_is_signal(self):
        s = hs.load(TEST_DATA_DIR / "20241021_00405_movie.mrc", lazy=True)
        assert isinstance(s.metadata._HyperSpy.navigator, hs.signals.BaseSignal)

    def test_virtual_image_is_signal2d(self):
        s = hs.load(TEST_DATA_DIR / "20241021_00405_movie.mrc", lazy=True)
        virt = s.metadata._HyperSpy.navigators["Virt 0"]
        assert isinstance(virt, hs.signals.Signal2D)


@pytest.mark.skipif(zarr_missing, reason="zarr not installed")
class TestLoadSave:
    @pytest.mark.parametrize(
        "metadata_file",
        [
            TEST_DATA_DIR / "3DSTEM_scan_info.txt",
            TEST_DATA_DIR / "3DTEM_scan_info.txt",
            TEST_DATA_DIR / "3DTEMDiffracting_scan_info.txt",
        ],
    )
    def test_mrc_metadata_save(self, metadata_file, tmp_path):
        """Saving an MRC-loaded signal to zspy should not raise a
        JSON-serialization error for numpy.void / bytes header fields."""
        s = hs.load(
            TEST_DATA_DIR / "20241021_00405_movie.mrc", metadata_file=metadata_file
        )
        diffracting = (
            "STEM" in metadata_file.name or "Diffracting" in metadata_file.name
        )
        s.axes_manager.navigation_axes[0].units = "sec"
        if diffracting:
            s.axes_manager.signal_axes[0].units = "nm^-1"
            s.axes_manager.signal_axes[1].units = "nm^-1"
        else:
            s.axes_manager.signal_axes[0].units = "nm"
            s.axes_manager.signal_axes[1].units = "nm"

        out = tmp_path / "out.zspy"
        # Must not raise TypeError: Object of type void is not JSON serializable
        s.save(str(out))

    def test_mrc_void_header_fields_are_hex_strings(self):
        """numpy.void and bytes fields in the MRC header (EXTRA, EXTRA2,
        CMAP, STAMP, LABELS) must be stored as plain hex strings, not raw
        numpy/bytes objects, so they can be JSON-serialized (e.g. for zspy)."""
        s = hs.load(TEST_DATA_DIR / "20241021_00405_movie.mrc")
        std_header = s.original_metadata.std_header
        for field in ("EXTRA", "EXTRA2", "CMAP", "STAMP", "LABELS"):
            value = std_header[field]
            assert isinstance(value, str), (
                f"Header field '{field}' should be a hex string, got {type(value)}"
            )
            # Must be a valid hex string
            bytes.fromhex(value)

    def test_mrc_save_reload_zspy(self, tmp_path):
        """Round-trip: load MRC, save to zspy, reload – data and key metadata
        should be preserved, and no serialization errors should occur."""
        zarr = pytest.importorskip("zarr", reason="zarr not installed")  # noqa: F841
        s = hs.load(TEST_DATA_DIR / "20241021_00405_movie.mrc", lazy=True)
        out = tmp_path / "round_trip.zspy"
        s.save(str(out))
        s2 = hs.load(str(out))
        np.testing.assert_array_equal(s.data.compute(), s2.data)



# -*- coding: utf-8 -*-
# Copyright 2024 The HyperSpy developers
#
# This file is part of rosettasciio.
#
# rosettasciio is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# rosettasciio is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with rosettasciio. If not, see <https://www.gnu.org/licenses/#GPL>.


import numpy as np
import pytest

from rsciio.tests.registry import TEST_DATA_REGISTRY

h5py = pytest.importorskip("h5py", reason="h5py not installed")
hdf5plugin = pytest.importorskip("hdf5plugin", reason="hdf5plugin not installed")


from rsciio.arina import file_reader  # noqa: E402


@pytest.fixture
def test_file():
    """Get the path to the test file."""
    return TEST_DATA_REGISTRY.fetch("arina/test_00_master.h5")


def test_file_reader(test_file):
    """Test basic file reading functionality."""
    result = file_reader(test_file)
    assert isinstance(result, list)
    assert len(result) == 1
    result = result[0]  # Get the first dictionary

    assert isinstance(result, dict)
    assert "data" in result
    assert "axes" in result
    assert "metadata" in result
    assert "original_metadata" in result
    assert "post_process" in result
    assert "mapping" in result

    # Check data structure
    assert isinstance(result["data"], np.ndarray)
    assert len(result["data"].shape) == 4

    # Check axes information
    assert len(result["axes"]) == 4
    for axis in result["axes"]:
        assert "name" in axis
        assert "scale" in axis
        assert "offset" in axis
        assert "units" in axis
        assert "size" in axis
        assert axis["offset"] == 0.0
        if axis["name"] in ["x", "y"]:
            assert axis["units"] == "A"
        else:
            assert axis["units"] == "1"


def test_file_reader_with_binning(test_file):
    """Test file reading with binning."""
    result = file_reader(test_file, rebin_diffraction=2)[0]
    assert isinstance(result["data"], np.ndarray)
    assert len(result["data"].shape) == 4
    # Check that the last two dimensions are halved
    original_shape = file_reader(test_file)[0]["data"].shape
    assert result["data"].shape[-2] == original_shape[-2] // 2
    assert result["data"].shape[-1] == original_shape[-1] // 2


def test_file_reader_with_dtype(test_file):
    """Test file reading with specified dataset dtype."""
    result = file_reader(test_file, dtype=np.float32)[0]
    assert result["data"].dtype == np.float32


def test_file_reader_with_flatfield(test_file):
    """Test file reading with flatfield correction."""
    # Create a simple flatfield
    with h5py.File(test_file, "r") as f:
        shape = f["entry/data/data_000001"].shape[1:]
    flatfield = np.ones(shape, dtype=np.float32)
    result = file_reader(test_file, flatfield=flatfield)[0]
    assert result["data"].dtype == np.float32


def test_file_reader_nonexistent_file():
    """Test file reading with nonexistent file."""
    with pytest.raises(FileNotFoundError):
        file_reader("nonexistent_file.h5")


def test_file_reader_lazy_not_implemented(test_file):
    """Test that lazy loading is not implemented."""
    with pytest.raises(NotImplementedError):
        file_reader(test_file, lazy=True)


def test_arina_reader_navigation_shape(test_file):
    """Test that navigation_shape parameter is correctly applied."""
    s = file_reader(test_file, navigation_shape=(4, 4))
    assert s[0]["data"].shape == (4, 4, 192, 192)


def test_file_reader_invalid_navigation_shape(test_file):
    """Test file reading with invalid navigation_shape."""
    # Get the total number of images
    with h5py.File(test_file, "r") as f:
        nimages = sum(f["entry/data"][dset].shape[0] for dset in f["entry/data"])
    # Use a scan width that doesn't divide evenly into the total number of images
    invalid_navigation_shape = nimages + 1
    with pytest.raises(ValueError):
        file_reader(test_file, navigation_shape=(invalid_navigation_shape, "auto"))



# -*- coding: utf-8 -*-
# Copyright 2007-2026 The HyperSpy developers
#
# This file is part of RosettaSciIO.
#
# RosettaSciIO is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# RosettaSciIO is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with RosettaSciIO. If not, see <https://www.gnu.org/licenses/#GPL>.


import functools
import inspect
import warnings
from typing import Callable, Optional, Union

from rsciio.exceptions import VisibleDeprecationWarning


class deprecated:
    """Decorator to mark deprecated functions with an informative
    warning.
    Adapted from
    `scikit-image
    <https://github.com/scikit-image/scikit-image/blob/main/skimage/_shared/utils.py>`_
    and `matplotlib
    <https://github.com/matplotlib/matplotlib/blob/master/lib/matplotlib/_api/deprecation.py>`_.
    """

    def __init__(
        self,
        since: Union[str, int, float],
        alternative: Optional[str] = None,
        alternative_is_function: bool = True,
        removal: Union[str, int, float, None] = None,
    ):
        """
        Visible deprecation warning.

        Parameters
        ----------
        since
            The release at which this API became deprecated.
        alternative
            An alternative API that the user may use in place of the
            deprecated API.
        alternative_is_function
            Whether the alternative is a function. Default is ``True``.
        removal
            The expected removal version.
        """
        self.since = since
        self.alternative = alternative
        self.alternative_is_function = alternative_is_function
        self.removal = removal

    def __call__(self, func: Callable):
        # Wrap function to raise warning when called, and add warning to
        # docstring
        if self.alternative is not None:
            if self.alternative_is_function:
                alt_msg = f" Use `{self.alternative}()` instead."
            else:
                alt_msg = f" Use `{self.alternative}` instead."
        else:
            alt_msg = ""
        if self.removal is not None:
            rm_msg = f" and will be removed in version {self.removal}"
        else:
            rm_msg = ""
        msg = f"Function `{func.__name__}()` is deprecated{rm_msg}.{alt_msg}"

        @functools.wraps(func)
        def wrapped(*args, **kwargs):
            warnings.simplefilter(
                action="always",
                category=VisibleDeprecationWarning,
                append=True,
            )
            func_code = func.__code__
            warnings.warn_explicit(
                message=msg,
                category=VisibleDeprecationWarning,
                filename=func_code.co_filename,
                lineno=func_code.co_firstlineno + 1,
            )
            return func(*args, **kwargs)

        # Modify docstring to display deprecation warning
        old_doc = inspect.cleandoc(func.__doc__ or "").strip("\n")
        notes_header = "\nNotes\n-----"
        new_doc = (
            f"[*Deprecated*] {old_doc}\n"
            f"{notes_header if notes_header not in old_doc else ''}\n"
            f".. deprecated:: {self.since}\n"
            f"   {msg.strip()}"  # Matplotlib uses three spaces
        )
        wrapped.__doc__ = new_doc

        return wrapped


class deprecated_argument:
    """
    Decorator to remove an argument from a function or method's signature.
    Adapted from `scikit-image
    <https://github.com/scikit-image/scikit-image/blob/main/skimage/_shared/utils.py>`_.
    """

    def __init__(
        self, name, since, removal=None, alternative=None, additional_msg=None
    ):
        self.name = name
        self.since = since
        self.removal = removal
        self.alternative = alternative
        self.additional_msg = additional_msg

    def __call__(self, func):
        @functools.wraps(func)
        def wrapped(*args, **kwargs):
            if self.name in kwargs.keys():
                msg = f"Argument `{self.name}` is deprecated"
                if self.removal is not None:
                    msg += f" and will be removed in version {self.removal}."
                else:
                    msg += " and has been removed. "
                    kwargs.pop(self.name)
                msg += f" To avoid this warning, please do not use `{self.name}`. "
                if self.alternative is not None:
                    msg += f"Use `{self.alternative}` instead. "
                    kwargs[self.alternative] = kwargs.pop(self.name)
                msg += f"See the documentation of `{func.__name__}()` for more details."
                if self.additional_msg is not None:
                    msg += f" {self.additional_msg}"
                warnings.simplefilter(
                    action="always", category=VisibleDeprecationWarning
                )
                func_code = func.__code__
                warnings.warn_explicit(
                    message=msg,
                    category=VisibleDeprecationWarning,
                    filename=func_code.co_filename,
                    lineno=func_code.co_firstlineno + 1,
                )
            return func(*args, **kwargs)

        return wrapped


def distributed_keyword_deprecation(func):
    return deprecated_argument(
        "distributed",
        "0.8.0",
        additional_msg=" Distributed memory mapping is now supported in the default implementation.",
    )(func)


def mmap_mode_keyword_deprecation(func):
    return deprecated_argument(
        "mmap_mode",
        "0.8.0",
        additional_msg=" `mmap_mode` is deprecated because it is not used anymore.",
    )(func)


def endianess_keyword_deprecation(func):
    return deprecated_argument(
        "endianess",  # codespell:ignore
        "0.13.0",
        alternative="endianness",
    )(func)



# -*- coding: utf-8 -*-
# Copyright 2007-2026 The HyperSpy developers
#
# This file is part of RosettaSciIO.
#
# RosettaSciIO is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# RosettaSciIO is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with RosettaSciIO. If not, see <https://www.gnu.org/licenses/#GPL>.
"""Utility functions for file handling."""


# ruff: noqa: F822

import importlib

__all__ = [
    "get_file_handle",
    "inspect_npy_bytes",
    "memmap_distributed",
]


def __dir__():
    return sorted(__all__)


_import_mapping = {
    "get_file_handle": "_tools",
    "inspect_npy_bytes": "_tools",
    "memmap_distributed": "_distributed",
}


def __getattr__(name):
    if name in __all__:
        submodule = _import_mapping[name]
        return getattr(importlib.import_module(f"rsciio.utils.{submodule}"), name)

    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")



# -*- coding: utf-8 -*-
# Copyright 2007-2026 The HyperSpy developers
#
# This file is part of RosettaSciIO.
#
# RosettaSciIO is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# RosettaSciIO is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with RosettaSciIO. If not, see <https://www.gnu.org/licenses/#GPL>.

from collections import OrderedDict

import numpy as np


def sarray2dict(sarray, dictionary=None):
    """Converts a struct array to an ordered dictionary

    Parameters
    ----------
    sarray: struct array
    dictionary: None or dict
        If dictionary is not None the content of sarray will be appended to the
        given dictionary

    Returns
    -------
    Ordered dictionary

    """
    if dictionary is None:
        dictionary = OrderedDict()
    for name in sarray.dtype.names:
        value = sarray[name][0] if len(sarray[name]) == 1 else sarray[name]
        dictionary[name] = value
    return dictionary


def dict2sarray(dictionary, sarray=None, dtype=None):
    """Populates a struct array from a dictionary

    Parameters
    ----------
    dictionary: dict
    sarray: struct array or None
        Either sarray or dtype must be given. If sarray is given, it is
        populated from the dictionary.
    dtype: None, numpy dtype or dtype list
        If sarray is None, dtype must be given. If so, a new struct array
        is created according to the dtype, which is then populated.

    Returns
    -------
    Structure array

    """
    if sarray is None:
        if dtype is None:
            raise ValueError("Either sarray or dtype need to be specified.")
        sarray = np.zeros((1,), dtype=dtype)
    for name in set(sarray.dtype.names).intersection(set(dictionary.keys())):
        if len(sarray[name]) == 1:
            sarray[name][0] = dictionary[name]
        else:
            sarray[name] = dictionary[name]
    return sarray


def is_dask_array(x):
    """Check if x is a dask array

    Parameters
    ----------
    x : object
        Object to be checked.
    """
    return hasattr(x, "__dask_keys__")



# -*- coding: utf-8 -*-
# Copyright 2007-2026 The HyperSpy developers
#
# This file is part of RosettaSciIO.
#
# RosettaSciIO is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# RosettaSciIO is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with RosettaSciIO. If not, see <https://www.gnu.org/licenses/#GPL>.

import logging

_logger = logging.getLogger(__name__)


def jit_ifnumba(*decorator_args, **decorator_kwargs):
    try:
        import numba

        decorator_kwargs.setdefault("nopython", True)
        return numba.jit(*decorator_args, **decorator_kwargs)
    except ImportError:
        _logger.warning(
            "Falling back to slow pure python code, because `numba` is not installed."
        )

        def wrap(func):
            def wrapper_func(*args, **kwargs):
                return func(*args, **kwargs)

            return wrapper_func

        return wrap



# -*- coding: utf-8 -*-
# Copyright 2007-2026 The HyperSpy developers
#
# This file is part of RosettaSciIO.
#
# RosettaSciIO is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# RosettaSciIO is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with RosettaSciIO. If not, see <https://www.gnu.org/licenses/#GPL>.


import os
from datetime import datetime

import numpy as np

from rsciio._docstrings import FILENAME_DOC, LAZY_UNSUPPORTED_DOC, RETURNS_DOC


def _cnv_time(timestr):
    try:
        if not isinstance(timestr, str):
            # for numpy < 2.0
            timestr = timestr.decode()
        t = datetime.strptime(timestr, "%H:%M:%S.%f")
        dt = t - datetime(t.year, t.month, t.day)
        r = float(dt.seconds) + float(dt.microseconds) * 1e-6
    except ValueError:
        r = float(timestr)
    return r


def _bad_file(filename):
    raise AssertionError("Cannot interpret as DENS heater log: %s" % filename)


def file_reader(filename, lazy=False):
    """
    Read a DENSsolutions DigiHeater logfile.

    Parameters
    ----------
    %s
    %s

    %s
    """
    if lazy is not False:
        raise NotImplementedError("Lazy loading is not supported.")

    with open(filename, "rt") as f:
        # Strip leading, empty lines
        line = str(f.readline())
        while line.strip() == "" and not f.closed:
            line = str(f.readline())
        try:
            date, version = line.split("\t")
        except ValueError:
            _bad_file(filename)
        if version.strip() != "Digiheater 3.1":
            _bad_file(filename)
        calib = str(f.readline()).split("\t")
        str(f.readline())  # delta_t
        header_line = str(f.readline())
        try:
            R0, a, b, c = [float(v.split("=")[1]) for v in calib]
            date0 = datetime.strptime(date, "%d/%m/'%y %H:%M")
            date = "%s" % date0.date()
            time = "%s" % date0.time()
        except ValueError:
            _bad_file(filename)
        original_metadata = dict(R0=R0, a=a, b=b, c=c, date=date0, version=version)

        if header_line.strip() != (
            "sample\ttime\tTset[C]\tTmeas[C]\tRheat[ohm]\tVheat[V]\t"
            "Iheat[mA]\tPheat [mW]\tc"
        ):
            _bad_file(filename)
        try:
            rawdata = np.loadtxt(
                f, converters={1: _cnv_time}, usecols=(1, 3), unpack=True
            )
        except ValueError:
            _bad_file(filename)

    times = rawdata[0]
    # Add a day worth of seconds to any values after a detected rollover
    # Hopefully unlikely that there is more than one, but we can handle it
    for rollover in 1 + np.where(np.diff(times) < 0)[0]:
        times[rollover:] += 60 * 60 * 24
    # Raw data is not necessarily grid aligned. Interpolate onto grid.
    offset, scale = np.polynomial.polynomial.polyfit(
        np.arange(times.size), times, deg=1
    )

    metadata = {
        "General": {
            "original_filename": os.path.split(filename)[1],
            "date": date,
            "time": time,
        },
        "Signal": {"signal_type": "", "quantity": "Temperature (Celsius)"},
    }

    axes = [
        {
            "size": times.size,
            "index_in_array": 0,
            "name": "Time",
            "scale": scale,
            "offset": offset,
            "units": "s",
            "navigate": False,
        }
    ]

    dictionary = {
        "data": rawdata[1],
        "axes": axes,
        "metadata": metadata,
        "original_metadata": {"DENS_header": original_metadata},
    }

    return [
        dictionary,
    ]


file_reader.__doc__ %= (FILENAME_DOC, LAZY_UNSUPPORTED_DOC, RETURNS_DOC)



# -*- coding: utf-8 -*-
# Copyright 2007-2026 The HyperSpy developers
#
# This file is part of RosettaSciIO.
#
# RosettaSciIO is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# RosettaSciIO is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with RosettaSciIO. If not, see <https://www.gnu.org/licenses/#GPL>.

import gc
import hashlib
import os
import tempfile
import zipfile
from pathlib import Path

import numpy as np
import pytest

from rsciio.edax import file_reader
from rsciio.utils._tests import expected_is_binned

hs = pytest.importorskip("hyperspy.api", reason="hyperspy not installed")
requests = pytest.importorskip("requests", reason="requests not installed")

TEST_DATA_PATH = Path(__file__).parent / "data" / "edax"
ZIPF = TEST_DATA_PATH / "edax_files.zip"
TMP_DIR = tempfile.TemporaryDirectory()
TEST_FILES_OK = os.path.isfile(ZIPF)
REASON = ""
SHA256SUM = "e217c71efbd208da4b52e9cf483443f9da2175f2924a96447ed393086fe32008"


# The test files are not included in HyperSpy v1.4 because their file size is 36.5MB
# taking the HyperSpy source distribution file size above PyPI's 60MB limit.
# As a temporary solution, we attempt to download the test files from GitHub
# and skip the tests if the download fails.
if not TEST_FILES_OK:
    try:
        r = requests.get(
            "https://github.com/hyperspy/hyperspy/blob/e7a323a3bb9b237c24bd9267d2cc4fcb31bb99f3/hyperspy/tests/io/edax_files.zip?raw=true"
        )

        SHA256SUM_GOT = hashlib.sha256(r.content).hexdigest()
        if SHA256SUM_GOT == SHA256SUM:
            with open(ZIPF, "wb") as f:
                f.write(r.content)
            TEST_FILES_OK = True
        else:
            REASON = (
                "wrong sha256sum of downloaded file. Expected: %s, got: %s" % SHA256SUM,
                SHA256SUM_GOT,
            )
    except Exception as e:
        REASON = "download of EDAX test files failed: %s" % e


def setup_module():
    if TEST_FILES_OK:
        with zipfile.ZipFile(ZIPF, "r") as zipped:
            zipped.extractall(TMP_DIR.name)


pytestmark = pytest.mark.skipif(not TEST_FILES_OK, reason=REASON)


def teardown_module():
    TMP_DIR.cleanup()


class TestSpcSpectrum_v061_xrf:
    @classmethod
    def setup_class(cls):
        cls.spc = hs.load(os.path.join(TMP_DIR.name, "spc0_61-ipr333_xrf.spc"))
        cls.spc_loadAll = hs.load(
            os.path.join(TMP_DIR.name, "spc0_61-ipr333_xrf.spc"), load_all_spc=True
        )

    @classmethod
    def teardown_class(cls):
        del cls.spc, cls.spc_loadAll
        gc.collect()

    def test_data(self):
        # test datatype
        assert np.uint32 == TestSpcSpectrum_v061_xrf.spc.data.dtype
        # test data shape
        assert (4000,) == TestSpcSpectrum_v061_xrf.spc.data.shape
        # test 40 datapoints
        assert [
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            319,
            504,
            639,
            924,
            1081,
            1326,
            1470,
            1727,
            1983,
            2123,
            2278,
            2509,
            2586,
            2639,
            2681,
            2833,
            2696,
            2704,
            2812,
            2745,
            2709,
            2647,
            2608,
            2620,
            2571,
            2669,
        ] == TestSpcSpectrum_v061_xrf.spc.data[:40].tolist()

    def test_parameters(self):
        elements = TestSpcSpectrum_v061_xrf.spc.metadata.as_dictionary()["Sample"][
            "elements"
        ]
        sem_dict = TestSpcSpectrum_v061_xrf.spc.metadata.as_dictionary()[
            "Acquisition_instrument"
        ]["SEM"]  # this will eventually need to
        #  be changed when XRF-specific
        #  features are added
        eds_dict = sem_dict["Detector"]["EDS"]
        signal_dict = TestSpcSpectrum_v061_xrf.spc.metadata.as_dictionary()["Signal"]

        # Testing SEM parameters
        np.testing.assert_allclose(30, sem_dict["beam_energy"])
        np.testing.assert_allclose(0, sem_dict["Stage"]["tilt_alpha"])

        # Testing EDS parameters
        np.testing.assert_allclose(45, eds_dict["azimuth_angle"])
        np.testing.assert_allclose(35, eds_dict["elevation_angle"])
        np.testing.assert_allclose(
            137.92946, eds_dict["energy_resolution_MnKa"], atol=1e-5
        )
        np.testing.assert_allclose(2561.0, eds_dict["live_time"], atol=1e-6)

        # Testing elements
        assert {"Al", "Ca", "Cl", "Cr", "Fe", "K", "Mg", "Mn", "Si", "Y"} == set(
            elements
        )

        # Testing HyperSpy parameters
        assert "EDS_SEM" == signal_dict["signal_type"]
        assert isinstance(TestSpcSpectrum_v061_xrf.spc, hs.signals.Signal1D)

    def test_axes(self):
        spc_ax_manager = {
            "axis-0": {
                "_type": "UniformDataAxis",
                "name": "Energy",
                "navigate": False,
                "is_binned": expected_is_binned(),
                "offset": 0.0,
                "scale": 0.01,
                "size": 4000,
                "units": "keV",
            }
        }
        assert (
            spc_ax_manager == TestSpcSpectrum_v061_xrf.spc.axes_manager.as_dictionary()
        )

    def test_load_all_spc(self):
        spc_header = TestSpcSpectrum_v061_xrf.spc_loadAll.original_metadata[
            "spc_header"
        ]

        np.testing.assert_allclose(4, spc_header["analysisType"])
        np.testing.assert_allclose(4, spc_header["analyzerType"])
        np.testing.assert_allclose(2013, spc_header["collectDateYear"])
        np.testing.assert_allclose(9, spc_header["collectDateMon"])
        np.testing.assert_allclose(26, spc_header["collectDateDay"])
        np.testing.assert_equal(b"Garnet1.", spc_header["fileName"].view("|S8")[0])
        np.testing.assert_allclose(45, spc_header["xRayTubeZ"])


class TestSpcSpectrum_v070_eds:
    @classmethod
    def setup_class(cls):
        cls.spc = hs.load(os.path.join(TMP_DIR.name, "single_spect.spc"))
        cls.spc_loadAll = hs.load(
            os.path.join(TMP_DIR.name, "single_spect.spc"), load_all_spc=True
        )

    @classmethod
    def teardown_class(cls):
        del cls.spc, cls.spc_loadAll
        gc.collect()

    def test_data(self):
        # test datatype
        assert np.uint32 == TestSpcSpectrum_v070_eds.spc.data.dtype
        # test data shape
        assert (4096,) == TestSpcSpectrum_v070_eds.spc.data.shape
        # test 1st 20 datapoints
        assert [
            0,
            0,
            0,
            0,
            0,
            0,
            1,
            2,
            3,
            3,
            10,
            4,
            10,
            10,
            45,
            87,
            146,
            236,
            312,
            342,
        ] == TestSpcSpectrum_v070_eds.spc.data[:20].tolist()

    def test_parameters(self):
        elements = TestSpcSpectrum_v070_eds.spc.metadata.as_dictionary()["Sample"][
            "elements"
        ]
        sem_dict = TestSpcSpectrum_v070_eds.spc.metadata.as_dictionary()[
            "Acquisition_instrument"
        ]["SEM"]
        eds_dict = sem_dict["Detector"]["EDS"]
        signal_dict = TestSpcSpectrum_v070_eds.spc.metadata.as_dictionary()["Signal"]

        # Testing SEM parameters
        np.testing.assert_allclose(22, sem_dict["beam_energy"])
        np.testing.assert_allclose(0, sem_dict["Stage"]["tilt_alpha"])

        # Testing EDS parameters
        np.testing.assert_allclose(0, eds_dict["azimuth_angle"])
        np.testing.assert_allclose(34, eds_dict["elevation_angle"])
        np.testing.assert_allclose(
            129.31299, eds_dict["energy_resolution_MnKa"], atol=1e-5
        )
        np.testing.assert_allclose(50.000004, eds_dict["live_time"], atol=1e-6)

        # Testing elements
        assert {"Al", "C", "Ce", "Cu", "F", "Ho", "Mg", "O"} == set(elements)

        # Testing HyperSpy parameters
        assert "EDS_SEM" == signal_dict["signal_type"]
        assert isinstance(TestSpcSpectrum_v070_eds.spc, hs.signals.Signal1D)

    def test_axes(self):
        spc_ax_manager = {
            "axis-0": {
                "_type": "UniformDataAxis",
                "name": "Energy",
                "navigate": False,
                "is_binned": expected_is_binned(),
                "offset": 0.0,
                "scale": 0.01,
                "size": 4096,
                "units": "keV",
            }
        }
        assert (
            spc_ax_manager == TestSpcSpectrum_v070_eds.spc.axes_manager.as_dictionary()
        )

    def test_load_all_spc(self):
        spc_header = TestSpcSpectrum_v070_eds.spc_loadAll.original_metadata[
            "spc_header"
        ]

        np.testing.assert_allclose(4, spc_header["analysisType"])
        np.testing.assert_allclose(5, spc_header["analyzerType"])
        np.testing.assert_allclose(2016, spc_header["collectDateYear"])
        np.testing.assert_allclose(4, spc_header["collectDateMon"])
        np.testing.assert_allclose(19, spc_header["collectDateDay"])
        np.testing.assert_equal(
            b"C:\\ProgramData\\EDAX\\jtaillon\\Cole\\Mapping\\Lsm\\"
            b"GFdCr\\950\\Area 1\\spectrum20160419153851427_0.spc",
            spc_header["longFileName"].view("|S256")[0],
        )
        np.testing.assert_allclose(0, spc_header["xRayTubeZ"])


class TestSpdMap_070_eds:
    @classmethod
    def setup_class(cls):
        cls.spd = hs.load(os.path.join(TMP_DIR.name, "spd_map.spd"), convert_units=True)

    @classmethod
    def teardown_class(cls):
        del cls.spd
        gc.collect()

    def test_data(self):
        # test d_type
        assert np.uint16 == TestSpdMap_070_eds.spd.data.dtype
        # test d_shape
        assert (200, 256, 2500) == TestSpdMap_070_eds.spd.data.shape
        assert (
            [
                [
                    [0, 0, 0, 0, 0],  # test random data
                    [0, 0, 1, 0, 1],
                    [0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0],
                ],
                [
                    [0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0],
                    [0, 0, 0, 1, 0],
                    [0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0],
                ],
                [
                    [0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 1],
                    [0, 1, 1, 0, 0],
                    [0, 0, 0, 0, 0],
                ],
                [
                    [0, 1, 0, 0, 0],
                    [0, 0, 0, 1, 0],
                    [0, 0, 0, 0, 0],
                    [0, 0, 1, 0, 0],
                    [0, 0, 0, 1, 0],
                ],
                [
                    [0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0],
                    [0, 0, 1, 0, 1],
                    [0, 0, 0, 1, 0],
                    [0, 0, 0, 0, 0],
                ],
            ]
            == TestSpdMap_070_eds.spd.data[15:20, 15:20, 15:20].tolist()
        )

    def test_parameters(self):
        elements = TestSpdMap_070_eds.spd.metadata.as_dictionary()["Sample"]["elements"]
        sem_dict = TestSpdMap_070_eds.spd.metadata.as_dictionary()[
            "Acquisition_instrument"
        ]["SEM"]
        eds_dict = sem_dict["Detector"]["EDS"]
        signal_dict = TestSpdMap_070_eds.spd.metadata.as_dictionary()["Signal"]

        # Testing SEM parameters
        np.testing.assert_allclose(22, sem_dict["beam_energy"])
        np.testing.assert_allclose(0, sem_dict["Stage"]["tilt_alpha"])

        # Testing EDS parameters
        np.testing.assert_allclose(0, eds_dict["azimuth_angle"])
        np.testing.assert_allclose(34, eds_dict["elevation_angle"])
        np.testing.assert_allclose(
            126.60252, eds_dict["energy_resolution_MnKa"], atol=1e-5
        )
        np.testing.assert_allclose(2621.4399, eds_dict["live_time"], atol=1e-4)

        # Testing elements
        assert {"Ce", "Co", "Cr", "Fe", "Gd", "La", "Mg", "O", "Sr"} == set(elements)

        # Testing HyperSpy parameters
        assert "EDS_SEM" == signal_dict["signal_type"]
        assert isinstance(TestSpdMap_070_eds.spd, hs.signals.Signal1D)

    def test_axes(self):
        spd_ax_manager = {
            "axis-0": {
                "_type": "UniformDataAxis",
                "name": "y",
                "navigate": True,
                "is_binned": False,
                "offset": 0.0,
                "scale": 14.227345585823057,
                "size": 200,
                "units": "nm",
            },
            "axis-1": {
                "_type": "UniformDataAxis",
                "name": "x",
                "navigate": True,
                "is_binned": False,
                "offset": 0.0,
                "scale": 14.235896058380602,
                "size": 256,
                "units": "nm",
            },
            "axis-2": {
                "_type": "UniformDataAxis",
                "name": "Energy",
                "navigate": False,
                "is_binned": expected_is_binned(),
                "offset": 0.0,
                "scale": 0.0050000000000000001,
                "size": 2500,
                "units": "keV",
            },
        }
        assert spd_ax_manager == TestSpdMap_070_eds.spd.axes_manager.as_dictionary()

    def test_ipr_reading(self):
        ipr_header = TestSpdMap_070_eds.spd.original_metadata["ipr_header"]
        np.testing.assert_allclose(0.014235896, ipr_header["mppX"])
        np.testing.assert_allclose(0.014227346, ipr_header["mppY"])

    def test_spc_reading(self):
        # Test to make sure that spc metadata matches spd metadata
        spc_header = TestSpdMap_070_eds.spd.original_metadata["spc_header"]

        elements = TestSpdMap_070_eds.spd.metadata.as_dictionary()["Sample"]["elements"]
        sem_dict = TestSpdMap_070_eds.spd.metadata.as_dictionary()[
            "Acquisition_instrument"
        ]["SEM"]
        eds_dict = sem_dict["Detector"]["EDS"]

        np.testing.assert_allclose(spc_header.azimuth, eds_dict["azimuth_angle"])
        np.testing.assert_allclose(
            spc_header.detReso, eds_dict["energy_resolution_MnKa"]
        )
        np.testing.assert_allclose(spc_header.elevation, eds_dict["elevation_angle"])
        np.testing.assert_allclose(spc_header.liveTime, eds_dict["live_time"])
        np.testing.assert_allclose(
            spc_header.evPerChan, TestSpdMap_070_eds.spd.axes_manager[2].scale * 1000
        )
        np.testing.assert_allclose(spc_header.kV, sem_dict["beam_energy"])
        np.testing.assert_allclose(spc_header.numElem, len(elements))


class TestSpdMap_061_xrf:
    @classmethod
    def setup_class(cls):
        cls.spd = hs.load(
            os.path.join(TMP_DIR.name, "spc0_61-ipr333_xrf.spd"), convert_units=True
        )

    @classmethod
    def teardown_class(cls):
        del cls.spd
        gc.collect()

    def test_data(self):
        # test d_type
        assert np.uint16 == TestSpdMap_061_xrf.spd.data.dtype
        # test d_shape
        assert (200, 256, 2000) == TestSpdMap_061_xrf.spd.data.shape
        assert [
            [
                [0, 0, 0, 0, 0],
                [0, 0, 0, 0, 0],
                [0, 0, 0, 0, 0],
                [0, 0, 0, 0, 0],
                [0, 0, 0, 1, 0],
            ],
            [
                [0, 0, 0, 0, 0],
                [0, 0, 0, 0, 0],
                [0, 0, 0, 0, 0],
                [0, 0, 0, 0, 0],
                [0, 0, 0, 0, 0],
            ],
            [
                [0, 0, 0, 0, 0],
                [0, 0, 0, 0, 0],
                [0, 0, 0, 0, 0],
                [0, 0, 0, 0, 0],
                [0, 0, 0, 0, 1],
            ],
            [
                [0, 0, 0, 0, 0],
                [0, 0, 0, 0, 0],
                [0, 0, 0, 0, 0],
                [0, 0, 0, 0, 0],
                [0, 0, 0, 0, 1],
            ],
            [
                [0, 0, 0, 0, 0],
                [0, 0, 0, 0, 0],
                [0, 0, 0, 0, 0],
                [0, 0, 0, 0, 0],
                [0, 0, 0, 0, 0],
            ],
        ] == TestSpdMap_061_xrf.spd.data[15:20, 15:20, 15:20].tolist()

    def test_parameters(self):
        elements = TestSpdMap_061_xrf.spd.metadata.as_dictionary()["Sample"]["elements"]
        sem_dict = TestSpdMap_061_xrf.spd.metadata.as_dictionary()[
            "Acquisition_instrument"
        ]["SEM"]
        eds_dict = sem_dict["Detector"]["EDS"]
        signal_dict = TestSpdMap_061_xrf.spd.metadata.as_dictionary()["Signal"]

        # Testing SEM parameters
        np.testing.assert_allclose(30, sem_dict["beam_energy"])
        np.testing.assert_allclose(0, sem_dict["Stage"]["tilt_alpha"])

        # Testing EDS parameters
        np.testing.assert_allclose(45, eds_dict["azimuth_angle"])
        np.testing.assert_allclose(35, eds_dict["elevation_angle"])
        np.testing.assert_allclose(
            137.92946, eds_dict["energy_resolution_MnKa"], atol=1e-5
        )
        np.testing.assert_allclose(2561.0, eds_dict["live_time"], atol=1e-4)

        # Testing elements
        assert {"Al", "Ca", "Cl", "Cr", "Fe", "K", "Mg", "Mn", "Si", "Y"} == set(
            elements
        )

        # Testing HyperSpy parameters
        assert "EDS_SEM" == signal_dict["signal_type"]
        assert isinstance(TestSpdMap_061_xrf.spd, hs.signals.Signal1D)

    def test_axes(self):
        spd_ax_manager = {
            "axis-0": {
                "_type": "UniformDataAxis",
                "name": "y",
                "navigate": True,
                "is_binned": False,
                "offset": 0.0,
                "scale": 0.5651920166015625,
                "size": 200,
                "units": "mm",
            },
            "axis-1": {
                "_type": "UniformDataAxis",
                "name": "x",
                "navigate": True,
                "is_binned": False,
                "offset": 0.0,
                "scale": 0.5651920166015625,
                "size": 256,
                "units": "mm",
            },
            "axis-2": {
                "_type": "UniformDataAxis",
                "name": "Energy",
                "navigate": False,
                "is_binned": expected_is_binned(),
                "offset": 0.0,
                "scale": 0.01,
                "size": 2000,
                "units": "keV",
            },
        }
        assert spd_ax_manager == TestSpdMap_061_xrf.spd.axes_manager.as_dictionary()

    def test_ipr_reading(self):
        ipr_header = TestSpdMap_061_xrf.spd.original_metadata["ipr_header"]
        np.testing.assert_allclose(565.1920166015625, ipr_header["mppX"])
        np.testing.assert_allclose(565.1920166015625, ipr_header["mppY"])

    def test_spc_reading(self):
        # Test to make sure that spc metadata matches spd_061_xrf metadata
        spc_header = TestSpdMap_061_xrf.spd.original_metadata["spc_header"]

        elements = TestSpdMap_061_xrf.spd.metadata.as_dictionary()["Sample"]["elements"]
        sem_dict = TestSpdMap_061_xrf.spd.metadata.as_dictionary()[
            "Acquisition_instrument"
        ]["SEM"]
        eds_dict = sem_dict["Detector"]["EDS"]

        np.testing.assert_allclose(spc_header.azimuth, eds_dict["azimuth_angle"])
        np.testing.assert_allclose(
            spc_header.detReso, eds_dict["energy_resolution_MnKa"]
        )
        np.testing.assert_allclose(spc_header.elevation, eds_dict["elevation_angle"])
        np.testing.assert_allclose(spc_header.liveTime, eds_dict["live_time"])
        np.testing.assert_allclose(
            spc_header.evPerChan, TestSpdMap_061_xrf.spd.axes_manager[2].scale * 1000
        )
        np.testing.assert_allclose(spc_header.kV, sem_dict["beam_energy"])
        np.testing.assert_allclose(spc_header.numElem, len(elements))


def test_unsupported_extension():
    with pytest.raises(ValueError):
        file_reader("fname.unsupported_extension")


@pytest.mark.parametrize("fname", ["single_spect.spc", "spd_map.spd"])
def test_lazy_loading(fname):
    spc = hs.load(os.path.join(TMP_DIR.name, fname))
    spc_lazy = hs.load(os.path.join(TMP_DIR.name, fname), lazy=True)
    assert spc_lazy._lazy

    spc_lazy.compute()

    np.testing.assert_allclose(spc_lazy.data, spc.data)



from pathlib import Path

import pytest

from rsciio.utils.xml import ET, XmlToDict, sanitize_msxml_float


@pytest.fixture
def XML_TEST_NODE():
    MY_PATH = Path(__file__).parent
    TEST_XML_PATH = MY_PATH / ".." / "data" / "ToastedBreakFastSDD.xml"

    with open(TEST_XML_PATH, "r") as fn:
        weird_but_valid_xml_str = fn.read()

    yield ET.fromstring(weird_but_valid_xml_str)


# fmt: off
def test_msxml_sanitization():
    bad_msxml = b"""
    <main>
      <simpleFloat>0,2</simpleFloat>
      <scientificSmall>1,9E-5</scientificSmall>
      <scientificBig>0,2E10</scientificBig>
      <actuallyList>0,2,3</actuallyList>
    </main>
    """
    sanitized_xml_bytes = sanitize_msxml_float(bad_msxml)
    et = ET.fromstring(sanitized_xml_bytes)
    assert et[0].text == "0.2"
    assert et[1].text == "1.9E-5"
    assert et[2].text == "0.2E10"
    assert et[3].text == "0,2,3"  # is not float


def test_default_x2d(XML_TEST_NODE):
    """test of default XmlToDict translation with attributes prefixed with @,
    interchild_text_parsing set to 'first',
    no flattening tags set, and dub_text_str set to '#value'
    """
    x2d = XmlToDict()
    pynode = x2d.dictionarize(XML_TEST_NODE)
    assert (
        pynode["TestXML"]["Main"]["ClassInstance"]["Sample"]["Components"][
            "ComponentChildren"
        ]["Instance"][0]["name"]
        == 'Eggs'
    )
    t = "With one of these components"
    assert pynode["TestXML"]["Main"]["ClassInstance"]["Sample"]["#value"] == t


def test_skip_interchild_text_flatten(XML_TEST_NODE):
    """test of XmlToDict translation with interchild_text_parsing set to 'skip',
    three string containing list set to flattening tags. Other kwrds - default.
    """
    x2d = XmlToDict(
        interchild_text_parsing='skip',
        tags_to_flatten=["ClassInstance", "ComponentChildren", "Instance", "TestXML"],
    )
    pynode = x2d.dictionarize(XML_TEST_NODE)
    assert pynode["Main"]["Sample"]["Components"]["name"][0] == "Eggs"
    assert pynode["Main"]["Sample"].get("#value") is None


def test_concat_interchild_text_val_flatten(XML_TEST_NODE):
    """test of XmlToDict translator with interchild_text_parsing set to
    'cat' (concatenation), four flattening tags set, and dub_text_str set
    to '#text'
    """
    x2d = XmlToDict(
        dub_text_str="#text",
        interchild_text_parsing='cat',
        tags_to_flatten=["ClassInstance", "ComponentChildren", "Instance", "Main"],
    )
    pynode = x2d.dictionarize(XML_TEST_NODE.find("Main"))
    assert pynode["Instrument"]["Type"].get("#value") is None
    assert pynode["Instrument"]["Type"]["#text"] == "Toaster"
    assert pynode["Sample"].get("#value") is None
    assert pynode["Sample"].get("#text") is None
    t = "With one of these componentsSDD risks to be Toasted."
    assert pynode["Sample"]["#interchild_text"] == t


def test_list_interchild_text_val_flatten(XML_TEST_NODE):
    """test of XmlToDict translator interchild_text_parsing set to 'list'
    """
    x2d = XmlToDict(
        dub_text_str="#value",
        interchild_text_parsing='list',
        tags_to_flatten=["ClassInstance", "ComponentChildren", "Instance"]
    )
    pynode = x2d.dictionarize(XML_TEST_NODE.find("Main//Sample"))
    assert pynode["Sample"]["#interchild_text"] == [
        "With one of these components",
        "",
        "SDD risks to be Toasted.",
    ]


def x2d_subclass_for_custom_bool(XML_TEST_NODE):
    """test subclass of XmlToDict with updated eval function"""

    class CustomXmlToDict(XmlToDict):
        @staticmethod
        def eval(string):
            if string == "not today":
                return False
            if string == "affirmative":
                return True
            return XmlToDict.eval(string)

    x2d = CustomXmlToDict(
        dub_text_str="#value",
        tags_to_flatten=["ClassInstance", "ComponentChildren", "Instance"],
    )
    node = XML_TEST_NODE.find("Main//Instrument")
    pynode = x2d.dictionarize(node)
    assert pynode["IsToasted"] is False
    assert pynode["IsToasting"] is True


def test_wrong_type_x2d_initiation():
    with pytest.raises(ValueError):
        XmlToDict(dub_attr_pre_str=1)
    with pytest.raises(ValueError):
        XmlToDict(tags_to_flatten=0)
    with pytest.raises(ValueError):
        XmlToDict(interchild_text_parsing="simple")
    with pytest.raises(ValueError):
        XmlToDict(dub_text_str=2)
